﻿B4J=true
Group=Classes\Modules
ModulesStructureVersion=1
Type=StaticCode
Version=10
@EndOfDesignText@
' EndsMeet Core
' Version 0.18
Sub Process_Globals
	Private su As StringUtils
	Private Pattern1() As String = Array As String("Root", "Path", "Controller", "First", "Second", "Third")
	Private Pattern2() As String = Array As String("Root", "Path", "Api", "Controller", "First", "Second", "Third")
	Private Pattern3() As String = Array As String("Root", "Path", "Api", "Version", "Controller", "First", "Second", "Third")
	Type Element (Index As Int, Name As String, Element As String)
	Type ControllerElement (Api As Element, Web As Element)
	Type BaseElements (Root As Element, Path As Element, Api As ApiElement, Web As WebElement, Controller As ControllerElement, First As Element, Second As Element, Third As Element)
	Type ApiElement (Index As Int, Name As String, Element As String, Controller As Element, Versioning As Boolean)
	Type WebElement (Index As Int, Name As String, Element As String, Controller As Element)
	Type BasicAuth (ClientId As String, ClientSecret As String)
	Type SimpleResponse (Enable As Boolean, Format As String, DataKey As String)
	Type HttpResponseMessage (ResponseCode As Int, ResponseString As String, ResponseData As List, ResponseObject As Map, ResponseMessage As String, ResponseError As Object, ResponseType As String, ContentType As String, SimpleResponse As SimpleResponse)
	Public HTML As String
	Public HTML404 As String = $"<h1>404 Page Not Found</h1>"$
	Public const CONTENT_TYPE_HTML As String = "text/html"
	Public const CONTENT_TYPE_JSON As String = "application/json"
End Sub

Public Sub RequestElements (Request As ServletRequest, BaseElements As BaseElements) As Map
	Dim Element() As String = Regex.Split("\/", Request.RequestURI)
	Dim PatternList As List = Pattern1
	Dim RequestElement As Map
	RequestElement.Initialize
	For i = 0 To Element.Length - 1
		If i = 2 Then
			If Element(i).EqualsIgnoreCase(BaseElements.Api.Element) Then
				If BaseElements.Api.Versioning Then
					Dim PatternList As List = Pattern3
				Else
					Dim PatternList As List = Pattern2
				End If
			End If
		End If
		RequestElement.Put(PatternList.Get(i), Element(i))
	Next
	Return RequestElement
End Sub

Public Sub isWebRequest (Request As ServletRequest, BaseElements As BaseElements) As Boolean
	Return RequestElements(Request, BaseElements).ContainsKey("Path")
End Sub

Public Sub isApiRequest (Request As ServletRequest, BaseElements As BaseElements) As Boolean
	Return RequestElements(Request, BaseElements).ContainsKey("Api")
End Sub

' Get Access Token from Header
Public Sub RequestAccessToken (req As ServletRequest) As String
	Dim token As String
	Dim auths As List = req.GetHeaders("Authorization")
	If auths.Size > 0 Then
		token = auths.Get(0)
	End If
	Return token
End Sub

Public Sub RequestBearerToken (req As ServletRequest) As String
	Dim Auths As List = req.GetHeaders("Authorization")
	If Auths.Size > 0 Then
		Dim auth As String = Auths.Get(0)
		If auth.StartsWith("Bearer") And auth.Length > "Bearer ".Length Then
			Return auth.SubString("Bearer ".Length)
		End If
	End If
	Return ""
End Sub

' Read Request Cookie As Map
Public Sub RequestCookie (req As ServletRequest) As Object
	Dim Munchies() As Cookie = req.GetCookies
	Dim M As Map
	M.Initialize
	For Each bite As Cookie In Munchies
		M.Put(bite.Name, bite.Value)
	Next
	Return M
End Sub

' Tip about POST requests: If you want to get a URL parameter (req.GetParameter)
' then do it only after reading the payload, otherwise the payload will be searched
' for the parameter and will be lost.
Public Sub RequestData (Request As ServletRequest) As String
	Dim inp As InputStream = Request.InputStream
	If inp.BytesAvailable <= 0 Then
		Return ""
	End If
	Dim buffer() As Byte = Bit.InputStreamToBytes(inp)
	Return BytesToString(buffer, 0, buffer.Length, "UTF-8")
End Sub

Public Sub RequestData2 (Request As ServletRequest) As Map
	Dim inp As InputStream = Request.InputStream
	If inp.BytesAvailable <= 0 Then
		Return CreateMap()
	End If
	Dim buffer() As Byte = Bit.InputStreamToBytes(inp)
	Dim str As String = BytesToString(buffer, 0, buffer.Length, "UTF-8")
	Return str.As(JSON).ToMap
End Sub

Public Sub RequestMultipartData (Request As ServletRequest) As Map
	Try
		Dim fd As String =  File.DirApp & "\www\tmp" ' & Main.inspectFolder
		Dim data As Map = Request.GetMultipartData(fd, 100000000 * 1000) 'max 10000000000 bytes
		For Each key As String In data.Keys
			Dim p As Part = data.Get(key)
			Dim name As String = p.SubmittedFilename 'data.Get("fn")
			Dim temp As String = File.GetName(p.TempFile)			
			Dim ins As InputStream = File.OpenInput(fd, temp)
			If key.StartsWith("post-") Then
				If File.Exists(fd, name) Then File.Delete(fd, name)
				Dim out As OutputStream = File.OpenOutput(fd, name, False)
				File.Copy2(ins, out)
				out.Close
				If File.Exists(fd, name) Then File.Delete(fd, temp) ' Delete temp file if new file generated
			End If
			If key.StartsWith("json") Then
				Dim buffer() As Byte = Bit.InputStreamToBytes(ins)
				Dim str As String = BytesToString(buffer, 0, buffer.Length, "UTF-8")
				data = str.As(JSON).ToMap
			End If
		Next
	Catch
		LogError(LastException.Message)
	End Try
	Return data
End Sub

Public Sub ReturnAuthorizationRequired (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 401
	mess.ResponseError = "Authentication required"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnBadRequest (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 400
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnErrorCredentialNotProvided (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 400
	mess.ResponseError = "Credential Not Provided"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnErrorExecuteQuery (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 422
	mess.ResponseError = "Execute Query"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnErrorUnprocessableEntity (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 422
	mess.ResponseError = "Unprocessable Entity"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnMethodNotAllow (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 405
	mess.ResponseError = "Method Not Allowed"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnTokenExpired (resp As ServletResponse)
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = 401
	mess.ResponseError = "Token Expired"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnHtmlPage (resp As ServletResponse)
	resp.ContentType = CONTENT_TYPE_HTML
	resp.Write(HTML)
End Sub

Public Sub ReturnHtmlPageNotFound (resp As ServletResponse)
	resp.ContentType = CONTENT_TYPE_HTML
	resp.Write(HTML404)
End Sub

Public Sub ReturnHttpResponse (mess As HttpResponseMessage, resp As ServletResponse)
	If mess.ResponseCode >= 200 And mess.ResponseCode < 300 Then ' SUCCESS
		If mess.ResponseType = "" Then mess.ResponseType = "SUCCESS"
		If mess.ResponseString = "" Then mess.ResponseString = "ok"
		If mess.ResponseMessage = "" Then mess.ResponseMessage = "Success"
		mess.ResponseError = Null
	Else ' ERROR
		If mess.ResponseCode = 0 Then mess.ResponseCode = 400
		If mess.ResponseType = "" Then mess.ResponseType = "ERROR"
		If mess.ResponseString = "" Then mess.ResponseString = "error"
		If mess.ResponseError.As(String).StartsWith("java.lang.Object") Then
			mess.ResponseError = "Bad Request"
			If mess.ResponseCode = 404 Then mess.ResponseError = "Not Found"
			If mess.ResponseCode = 405 Then mess.ResponseError = "Method Not Allowed"
			If mess.ResponseCode = 422 Then mess.ResponseError = "Unprocessable Entity"
		End If
	End If
	If mess.ContentType = "" Then mess.ContentType = CONTENT_TYPE_JSON
	resp.ContentType = mess.ContentType
	
	If mess.SimpleResponse.Enable Then
		resp.Status = mess.ResponseCode
	Else
		' Override Status Code
		If mess.ResponseCode < 200 Or mess.ResponseCode > 299 Then
			resp.Status = 200
		Else
			resp.Status = mess.ResponseCode
		End If
	End If

	If mess.SimpleResponse.Enable Then
		Select mess.SimpleResponse.Format
			Case "List"
				' force List format
				If mess.ResponseData.IsInitialized = False Then
					' Initialize an empty list
					mess.ResponseData.Initialize
					If mess.ResponseObject.IsInitialized Then
						' Add object to List
						If mess.ResponseObject.Size > 0 Then mess.ResponseData.Add(mess.ResponseObject)
					Else
						mess.ResponseObject.Initialize
					End If
				End If
				resp.Write(mess.ResponseData.As(JSON).ToString)
			Case "Map"
				' force Map format with "data" as key
				If mess.ResponseObject.IsInitialized = False Then
					' Initialize an empty map
					mess.ResponseObject.Initialize
					If mess.ResponseData.IsInitialized = False Then
						' Initialize an empty list
						mess.ResponseData.Initialize
					End If
					' Add object to Map
					If mess.SimpleResponse.DataKey = Null Then
						mess.ResponseObject.Put("data", mess.ResponseData)
					Else
						mess.ResponseObject.Put(mess.SimpleResponse.DataKey, mess.ResponseData)
					End If
				End If
				resp.Write(mess.ResponseObject.As(JSON).ToString)
			Case Else ' "Auto" or other value
				' Depends on which object is initialized
				If mess.ResponseObject.IsInitialized Then
					resp.Write(mess.ResponseObject.As(JSON).ToString)
				Else If mess.ResponseData.IsInitialized Then
					resp.Write(mess.ResponseData.As(JSON).ToString)
				Else
					' Default to an initialized List
					'mess.ResponseData.Initialize
					'resp.Write(mess.ResponseData.As(JSON).ToString)
					' or
					' Default to an initialized Map
					mess.ResponseObject.Initialize
					Select mess.ResponseType.ToUpperCase
						Case "SUCCESS"
							mess.ResponseObject.Put("message", mess.ResponseMessage)
						Case "ERROR", "FAILED"
							mess.ResponseObject.Put("error", mess.ResponseError)
					End Select
					resp.Write(mess.ResponseObject.As(JSON).ToString)
				End If
		End Select
	Else
		If Not(mess.ResponseData.IsInitialized) Then
			mess.ResponseData.Initialize
			If mess.ResponseObject.IsInitialized Then
				' Do not add an empty map
				If mess.ResponseObject.Size > 0 Then mess.ResponseData.Add(mess.ResponseObject)
			End If
		End If
		Dim map As Map = CreateMap("s": mess.ResponseString, "a": mess.ResponseCode, "r": mess.ResponseData, "m": mess.ResponseMessage, "e": mess.ResponseError)
		resp.Write(map.As(JSON).ToString)
	End If
End Sub

Public Sub ReturnCookie (key As String, value As String, max_age As Int, http_only As Boolean, resp As ServletResponse)
	Dim session_cookie As Cookie
	session_cookie.Initialize(key, value)
	session_cookie.HttpOnly = http_only
	session_cookie.MaxAge = max_age
	resp.AddCookie(session_cookie)
End Sub

Public Sub ReturnError (Code As Int, Error As String, resp As ServletResponse)
	If Code = 0 Then Code = 400
	If Error = "" Then Error = "Bad Request"
	Dim mess As HttpResponseMessage
	mess.Initialize
	mess.ResponseCode = Code
	mess.ResponseError = Error
	mess.ResponseString = "error"
	mess.SimpleResponse = Main.SimpleResponse
	ReturnHttpResponse(mess, resp)
End Sub

Public Sub ReturnLocation (Location As String, resp As ServletResponse) ' Code = 302
	resp.SendRedirect(Location)
End Sub


Public Sub BuildHtml (strHTML As String, Settings As Map) As String
	' Replace $KEY$ tag with new content from Map
	strHTML = ReplaceMap(strHTML, Settings)
	Return strHTML
End Sub

Public Sub BuildView (strHTML As String, View As String) As String
	' Replace @VIEW@ tag with new content
	strHTML = strHTML.Replace("@VIEW@", View)
	Return strHTML
End Sub

Public Sub BuildCsrfToken (strHTML As String, Content As String) As String
	' Replace meta name="csrf-token" tag with new content
	Dim strMetaTag As String = $"<meta name="csrf-token" content="${Content}">"$
	strHTML = strHTML.Replace($"<meta name="csrf-token" content="">"$, strMetaTag)
	Return strHTML
End Sub

Public Sub BuildTag (strHTML As String, Tag As String, Value As String) As String
	' Replace @TAG@ keyword with new content
	strHTML = strHTML.Replace("@" & Tag & "@", Value)
	Return strHTML
End Sub

Public Sub BuildDocView (strHTML As String, View As String) As String
	' Replace @DOCVIEW@ tag with new content
	strHTML = strHTML.Replace("@DOCVIEW@", View)
	Return strHTML
End Sub

Public Sub BuildScript (strHTML As String, Script As String) As String
	' Replace @SCRIPT@ tag with new content
	strHTML = strHTML.Replace("@SCRIPT@", Script)
	Return strHTML
End Sub

' Insert inline JavaScript
Public Sub BuildScript2 (strHTML As String, Script As String, Settings As Map) As String
	' Replace @SCRIPT@ tag with new content
	Dim strScript As String = $"	<script type="text/javascript">
		${Script}
	</script>"$
	strScript = BuildHtml(strScript, Settings)
	strHTML = strHTML.Replace("@SCRIPT@", strScript)
	Return strHTML
End Sub

Public Sub CheckAllowedVerb (SupportedMethods As List, Method As String) As Boolean
	'Methods: POST, GET, PUT, PATCH, DELETE
	If SupportedMethods.IndexOf(Method) = -1 Then
		Return False
	End If
	Return True
End Sub

'Deprecated
Public Sub CheckInteger (Input As Object) As Boolean
	Try
		Return Input > -1
	Catch
		'Log(LastException.Message)
		Return False
	End Try
End Sub

'Deprecated
Public Sub CheckMaxElements (Elements() As String, Max_Elements As Int) As Boolean
	If Elements.Length > Max_Elements Or Elements.Length = 0 Then
		Return False
	End If
	Return True
End Sub

Public Sub CurrentDateTime As String
	Dim CurrentDateFormat As String = DateTime.DateFormat
	DateTime.DateFormat = "yyyy-MM-dd HH:mm:ss"
	DateTime.SetTimeZone(0)
	Dim Current As String = DateTime.Date(DateTime.Now)
	DateTime.DateFormat = CurrentDateFormat
	Return Current
End Sub

Public Sub CurrentTimezone As String
	Return DateTime.GetTimeZoneOffsetAt(DateTime.Now)
End Sub

Public Sub DecodeBase64 (str As String) As Byte()
	Return su.DecodeBase64(str)
End Sub

Public Sub EncodeBase64 (data() As Byte) As String
	Return su.EncodeBase64(data)
End Sub

Public Sub DecodeBasicAuth (Auths As List) As BasicAuth
	Dim BasicAuth As BasicAuth
	BasicAuth.Initialize
	If Auths.Size > 0 Then
		Dim auth As String = Auths.Get(0)
		If auth.StartsWith("Basic") Then
			Dim b() As Byte = DecodeBase64(auth.SubString("Basic ".Length))
			Dim raw As String = BytesToString(b, 0, b.Length, "utf8")
			Dim UsernameAndPassword() As String = Regex.Split(":", raw)
			If UsernameAndPassword.Length = 2 Then
				BasicAuth.ClientId = UsernameAndPassword(0)
				BasicAuth.ClientSecret = UsernameAndPassword(1)
			End If
		End If
	End If
	Return BasicAuth
End Sub

Public Sub DecodeURL (str As String) As String
	Return su.DecodeUrl(str, "UTF8")
End Sub

Public Sub EncodeURL (str As String) As String
	Return su.EncodeUrl(str, "UTF8")
End Sub

' Taken from WebUtils
Public Sub EscapeHtml (Raw As String) As String
	Dim sb As StringBuilder
	sb.Initialize
	For i = 0 To Raw.Length - 1
		Dim C As Char = Raw.CharAt(i)
		Select C
			Case QUOTE
				sb.Append("&quot;")
			Case "'"
				sb.Append("&#39;")
			Case "<"
				sb.Append("&lt;")
			Case ">"
				sb.Append("&gt;")
			Case "&"
				sb.Append("&amp;")
			Case Else
				sb.Append(C)
		End Select
	Next
	Return sb.ToString
End Sub

Public Sub FileNameByCurrentDateTime As String
	Dim CurrentDateFormat As String = DateTime.DateFormat
	DateTime.DateFormat = "yyyyMMddHHmmss"
	DateTime.SetTimeZone(0)
	Dim Current As String = DateTime.Date(DateTime.Now)
	DateTime.DateFormat = CurrentDateFormat
	Return Current
End Sub

'Deprecated
Public Sub GetApiPathElements (Path As String) As String()
	Dim element() As String = Regex.Split("\/", Path)
	Return element
End Sub

'Deprecated
Public Sub GetUriElements (Uri As String) As String()
	Dim element() As String = Regex.Split("\/", Uri)
	Return element
End Sub

Public Sub GUID As String
	Dim sb As StringBuilder
	sb.Initialize
	For Each stp As Int In Array(8, 4, 4, 4, 12)
		If sb.Length > 0 Then sb.Append("-")
		For n = 1 To stp
			Dim c As Int = Rnd(0, 16)
			If c < 10 Then c = c + 48 Else c = c + 55
			sb.Append(Chr(c))
		Next
	Next
	Return sb.ToString
End Sub

Public Sub ProperCase (Word As String) As String
	If Word.Length = 0 Then Return ""
	If Word.Length = 1 Then Return Word.ToUpperCase
	Return Word.CharAt(0).As(String).ToUpperCase & Word.SubString(1)
End Sub

Public Sub ReadTextFile (FileName As String) As String
	Return File.ReadString(File.DirAssets, FileName)
End Sub

Public Sub WriteTextFile (FileName As String, Contents As String)
	File.WriteString(File.DirApp, FileName, Contents)
End Sub

'Public Sub AssetFileExist (FileName As String) As Boolean
'	Return File.Exists(File.Combine(File.DirApp, "www"), FileName)
'End Sub

Public Sub WriteAssetFile (FileName As String, Contents As String)
	File.WriteString(File.Combine(File.DirApp, "www"), FileName, Contents)
End Sub

Public Sub DeleteAssetFile (FileName As String)
	File.Delete(File.Combine(File.DirApp, "www"), FileName)
End Sub

' Taken from WebUtils
Public Sub ReplaceMap (Base As String, Replacements As Map) As String
	Dim pattern As StringBuilder
	pattern.Initialize
	For Each k As String In Replacements.Keys
		If pattern.Length > 0 Then pattern.Append("|")
		pattern.Append("\$").Append(k).Append("\$")
	Next
	Dim m As Matcher = Regex.Matcher(pattern.ToString, Base)
	Dim result As StringBuilder
	result.Initialize
	Dim lastIndex As Int
	Do While m.Find
		result.Append(Base.SubString2(lastIndex, m.GetStart(0)))
		Dim replace As String = Replacements.Get(m.Match.SubString2(1, m.Match.Length - 1))
		If m.Match.ToLowerCase.StartsWith("$h_") Then replace = EscapeHtml(replace)
		result.Append(replace)
		lastIndex = m.GetEnd(0)
	Loop
	If lastIndex < Base.Length Then result.Append(Base.SubString(lastIndex))
	Return result.ToString
End Sub

Public Sub Slugify (str As String) As String
	str = str.ToLowerCase.Trim
	str = Regex.Replace("/[^\w\s-]/g", str, "")
	str = Regex.Replace("/[\s_-]+/g", str, "-")
	str = Regex.Replace("/^-+|-+$/g", str, "")
	Return str
End Sub

' Check anti csrf-token variable sent from client in request header is same as variable stored in server session  
Public Sub ValidateCsrfToken (session_name As String, header_name As String, req As ServletRequest) As Boolean
	Dim headers As List = req.GetHeaders(header_name)
	If req.GetSession.GetAttribute2(session_name, "").As(String).EqualsIgnoreCase(headers.Get(0)) Then
		Return True
	Else
		Return False
	End If
End Sub

' // Source: http://www.b4x.com/android/forum/threads/validate-a-correctly-formatted-email-address.39803/
Public Sub ValidateEmail (EmailAddress As String) As Boolean
	Dim MatchEmail As Matcher = Regex.Matcher("^(?i)[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])$", EmailAddress)
 
	If MatchEmail.Find = True Then
		'Log(MatchEmail.Match)
		Return True
	Else
		'Log("Oops, please double check your email address...")
		Return False
	End If
End Sub